"""Simple polygon builder methods."""

import math
from typing import TYPE_CHECKING, Iterable, List

import shapely.wkt
from shapely.geometry import Point, Polygon

if TYPE_CHECKING:  # pragma: no cover
    from nzshm_common import CodedLocation


def create_hexagon(edge: float, x: float, y: float):
    """
    Create a hexagon centered on (x, y)
    :param edge: length of the hexagon's edge
    :param x: x-coordinate of the hexagon's center
    :param y: y-coordinate of the hexagon's center
    :return: The polygon containing the hexagon's coordinates
    """
    c = [
        [x + math.cos(math.radians(angle)) * edge, y + math.sin(math.radians(angle)) * edge]
        for angle in range(0, 360, 60)
    ]
    return Polygon(c)


def create_square_tile(dim: float, x: float, y: float):
    """
    Create a tile of size dim*dim, centered on (x, y)
    :param dim: length of the tiles edges
    :param x: x-coordinate of the tile's center
    :param y: y-coordinate of the tile's center
    :return: The polygon
    """
    offset = dim / 2
    c = [
        (x + offset, y + offset),
        (x + offset, y - offset),
        (x - offset, y - offset),
        (x - offset, y + offset),
        (x + offset, y + offset),
    ]
    return Polygon(c)


BA_POLYGON_WKT = (
    "POLYGON ((177.2 -37.715, 176.2 -38.72, 175.375 -39.27, "
    "174.25 -40, 173.1 -39.183, 171.7 -34.76, 173.54 -33.22, 177.2 -37.715))"
)


def backarc_polygon() -> Polygon:
    """
    Retrieve the backarc polygon from json and return shapely Polygon object
    """

    return shapely.wkt.loads(BA_POLYGON_WKT)


# TODO: consider if this function and any of the geometry operation functions belong in thier own repo. This
# would remove dependencies from what is supposed to be a pure python library and allow us to consolodate
# all geometry operations into one lib (including those in solvis and eq-fault-geom)
def within_polygon(locations: Iterable['CodedLocation'], polygon: Polygon) -> List[bool]:
    """
    Check if points are within a given polygon.
    Uses shapley.geometry.Polygon.contains() which will be false for points on the polygon boundary.

    Args:
        locations: the points to check
        polygon: the polygon

    Returns:
        A list of boolian values True if the point is within the polygon and False if not
    """

    points = [Point(loc.lon, loc.lat) for loc in locations]
    return [polygon.contains(point) for point in points]



__author__ = "GNS Science"
__email__ = 'nshm@gns.cri.nz'
__version__ = '0.8.4'


# Common classes at the top level for convenience
from .location.coded_location import CodedLocation, CodedLocationBin
from .location.types import LatLon



from pathlib import Path

import pytest

from nzshm_common.location.coded_location import CodedLocation
from nzshm_common.location.location import LOCATION_LISTS, get_locations

LOCATIONS_FILEPATH = Path(__file__).parent / 'fixtures' / 'location_file.csv'


def test_id():
    expected = [
        CodedLocation(lat=-41.3, lon=174.78, resolution=0.001),
        CodedLocation(lat=-36.87, lon=174.77, resolution=0.001),
    ]
    assert get_locations(["WLG", "AKL"]) == expected


def test_list():
    assert len(get_locations(["NZ", "SRWG214"])) == (
        len(LOCATION_LISTS["NZ"]["locations"]) + len(LOCATION_LISTS["SRWG214"]["locations"])
    )


def test_csv():
    expected = [
        CodedLocation(-41.2, 100.2, 0.001),
        CodedLocation(-30.5, 99, 0.001),
    ]
    assert get_locations([LOCATIONS_FILEPATH]) == expected


def test_mix():
    assert len(get_locations(["NZ", LOCATIONS_FILEPATH])) == 2 + len(LOCATION_LISTS["NZ"]["locations"])


def test_code():
    """Test for lat~lon code format."""
    expected = [
        CodedLocation(-41.2, 100.2, 0.001),
        CodedLocation(-30.5, 99, 0.001),
    ]
    assert (
        get_locations(["-41.200~100.200", "-30.500~99.000"]) == expected
    ), "Should work with codes at default resolution"
    assert (
        get_locations(["-41.2~100.2", "-30.5~99"]) == expected
    ), "Should work when code resolution doesn't exactly match, as long as values are correct"


def test_missing_location_name():
    with pytest.raises(KeyError, match="location missing_name is not a valid location identifier"):
        get_locations(["missing_name"])



import math
import unittest

from nzshm_common import CodedLocation
from nzshm_common.grids import RegionGrid, load_grid

try:
    import shapely  # noqa

    from nzshm_common.geometry.geometry import (  # these require shapely
        create_hexagon,
        create_square_tile,
        within_polygon,
    )

    HAVE_SHAPELY = True
except ImportError:
    HAVE_SHAPELY = False


class TestSquareTile(unittest.TestCase):
    @unittest.skipUnless(HAVE_SHAPELY, "Test requires optional shapely module.")
    def test_adjacent_tiles_pt1(self):
        """Get some 0.1 degree locations, check we can build tiles."""

        RESOLUTION = 0.1  # degrees
        grid = RegionGrid['NZ_0_1_NB_1_1']

        cells = []
        for pt in grid.load()[30:33]:
            cells.append(create_square_tile(RESOLUTION, *pt))

        for cell in cells:
            print(cell.centroid, cell.bounds[2] - cell.bounds[0])

        print('centroid spacing', cells[0].centroid.distance(cells[1].centroid))
        print('cell center to boundary', cells[0].centroid.distance(cells[0].boundary))
        print('cell spacing', cells[0].distance(cells[1]))
        print('cell exterior', list(cells[0].exterior.coords))
        self.assertAlmostEqual(cells[0].distance(cells[1]), 0.0)


# TODO: need a different grid layout...
def create_hexgrid(bbox, side):
    """
    returns an array of Points describing hexagons centers that are inside the given bounding_box
    :param bbox: The containing bounding box. The bbox coordinate should be in Webmercator.
    :param side: The size of the hexagons'
    :return: The hexagon grid
    """
    pass


class TestHex(unittest.TestCase):

    unittest.skipUnless(HAVE_SHAPELY, "Test requires optional shapely module.")

    def test_create_hexagon(self):
        RESOLUTION = 1  # degrees
        edge = math.sqrt(RESOLUTION**2 / (3 / 2 * math.sqrt(3)))

        hex = create_hexagon(edge, x=-45, y=175)
        print(hex)
        self.assertAlmostEqual(-45, hex.centroid.bounds[0])  # latitude
        self.assertAlmostEqual(175, hex.centroid.bounds[1])  # longiutude

    unittest.skipUnless(HAVE_SHAPELY, "Test requires optional shapely module.")

    def test_adjacent_hexagons_pt2(self):
        """Get some 0.2 degree locations, check we can buld adjacent hexagons."""

        RESOLUTION = 0.2  # degrees
        edge = math.sqrt(RESOLUTION**2 / (3 / 2 * math.sqrt(3)))

        print(f'edge length {edge}')
        grid = load_grid('NZ_0_2_NB_1_1')

        cells = []
        for pt in grid[30:33]:
            cells.append(create_hexagon(edge, *pt))

        for cell in cells:
            print(cell.centroid, cell.bounds[2] - cell.bounds[0])

        print('centroid spacing', cells[0].centroid.distance(cells[1].centroid))
        print('cell center to boundary', cells[0].centroid.distance(cells[0].boundary))
        print('cell spacing', cells[0].distance(cells[1]))

        self.assertEqual(cells[0].distance(cells[1]), 0.0)

    unittest.skipUnless(HAVE_SHAPELY, "Test requires optional shapely module.")

    def test_adjacent_hexagons_pt1(self):
        """Get some 0.1 degree locations, check we can buld adjacent hexagons."""

        RESOLUTION = 0.1  # degrees
        edge = math.sqrt(RESOLUTION**2 / (3 / 2 * math.sqrt(3)))

        print(f'edge length {edge}')
        grid = load_grid('NZ_0_1_NB_1_1')

        cells = []
        for pt in grid[30:33]:
            cells.append(create_hexagon(edge, *pt))

        for cell in cells:
            print(cell.centroid, cell.bounds[2] - cell.bounds[0])

        print('centroid spacing', cells[0].centroid.distance(cells[1].centroid))
        print('cell center to boundary', cells[0].centroid.distance(cells[0].boundary))
        print('cell spacing', cells[0].distance(cells[1]))

        self.assertAlmostEqual(cells[0].distance(cells[1]), 0.0)


@unittest.skipUnless(HAVE_SHAPELY, "Test requires optional shapely module.")
def test_within_polygon():
    coords = ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0))
    polygon = shapely.geometry.Polygon(coords)
    locations = [
        CodedLocation(0, 0, 0.001),
        CodedLocation(2, 2, 0.001),
        CodedLocation(0.5, 2, 0.001),
        CodedLocation(0.5, 0.5, 0.001),
    ]
    within_expected = [False, False, False, True]

    within = within_polygon(locations, polygon)

    assert len(within) == len(within_expected)
    assert within == within_expected



from typing import Iterable

import pytest

from nzshm_common.grids import RegionGrid, get_location_grid, get_location_grid_names, load_grid
from nzshm_common.location import CodedLocation


def test_load_wlg_0_005():
    assert len(load_grid('WLG_0_05_nb_1_1')) == 62


def test_load_wlg_0_001():
    assert len(load_grid('WLG_0_01_nb_1_1')) == 764


def test_load_nz_0_1():
    assert len(load_grid('NZ_0_1_NB_1_1')) == 3741


def test_load_lat_lon_order_spacing():
    """Coordinate order must be lat, lon."""
    grid = load_grid('NZ_0_1_NB_1_0')
    assert grid[0] == (-46.1, 166.4)
    assert grid[1] == (-46.0, 166.4)

    grid = load_grid('NZ_0_1_NB_1_1')
    assert grid[0] == (-46.1, 166.4)
    assert grid[1] == (-46.0, 166.4)

    grid = load_grid('NZ_0_2_NB_1_1')
    assert grid[0] == (-46.4, 166.6)
    assert grid[1] == (-46.2, 166.6)

    grid = load_grid('WLG_0_05_nb_1_1')
    assert grid[0] == (-41.4, 174.65)
    assert grid[1] == (-41.35, 174.65)
    assert grid[2] == (-41.3, 174.65)

    grid = load_grid('WLG_0_01_nb_1_1')
    assert grid[0] == (-41.36, 174.69)
    assert grid[1] == (-41.35, 174.69)
    assert grid[2] == (-41.34, 174.69)


def test_get_location_grid_default():
    """Test get_location_grid with standard default resolution."""
    # LatLons for comparison
    baseline_grid = load_grid("WLG_0_05_nb_1_1")

    grid_list = get_location_grid("WLG_0_05_nb_1_1")

    assert isinstance(grid_list, Iterable), "Should be Iterable"
    assert isinstance(grid_list[0], CodedLocation), "Should contain CodedLocation values"
    assert grid_list[0].resolution == RegionGrid.WLG_0_05_nb_1_1.resolution, "Should have grid default resolution"

    assert len(grid_list) == len(baseline_grid), "Should have a CodedLocation for each grid coordinate"
    assert grid_list[0].as_tuple == baseline_grid[0], "Preserving grid ordering"
    assert grid_list[1].as_tuple == baseline_grid[1], "Preserving grid ordering"
    assert grid_list[2].as_tuple == baseline_grid[2], "Preserving grid ordering"


def test_get_location_grid_downsampling():
    """Test behaviours of get_location_grid when setting a lower resolution."""
    # Load at default resolution for comparison
    grid_list = get_location_grid("WLG_0_05_nb_1_1")

    expected_message = "The requested resolution is lower than the grid resolution and will result in fewer points."

    with pytest.warns(UserWarning) as warnings:
        grid_downsampled = get_location_grid("WLG_0_05_nb_1_1", resolution=0.1)

    assert isinstance(grid_downsampled, Iterable), "Should be Iterable"
    assert isinstance(grid_downsampled[0], CodedLocation), "Should contain CodedLocation values"
    assert len(grid_list) > len(grid_downsampled), "Should have fewer downsampled coordinates"

    assert grid_list[0].as_tuple == (-41.4, 174.65), "Baseline value 0"
    # -- Skipped duplicated grid point --
    assert grid_list[2].as_tuple == (-41.3, 174.65), "Baseline value 2"
    # -- Skipped duplicated grid point --
    assert grid_list[4].as_tuple == (-41.35, 174.7), "Baseline value 4"

    assert grid_downsampled[0].as_tuple == (-41.4, 174.6), "Downsampled value 0"
    assert grid_downsampled[1].as_tuple == (-41.3, 174.6), "Downsampled value 1"
    assert grid_downsampled[2].as_tuple == (-41.4, 174.7), "Downsampled value 2"

    resample = grid_downsampled[0].resample(0.001)
    assert grid_downsampled[0] != resample, "Same locations at different resolutions are not considered equal"
    assert (
        grid_downsampled[0].as_tuple == resample.as_tuple
    ), "CodedLocations at different resolutions can be compared as LatLon values"

    assert len(warnings) == 1, "Should only be warned about downsampling"
    assert str(warnings[0].message) == expected_message, "Should have matched expected message"
    assert warnings[0].filename == __file__, "Should have been attributed to this caller"


def test_get_location_grid_names():
    name_list = get_location_grid_names()
    assert isinstance(name_list, Iterable), "Should be Iterable type"
    assert "NZ_0_1_NB_1_0" in name_list, "Should include NZ_... grids"
    assert "WLG_0_01_nb_1_1" in name_list, "Should include WLG_... grids"
    assert "SRWG214" not in name_list, "Should not include location list names"



"""
This module contains classes and functions for **coded locations**:
resolving map coordinates to a specific grid resolution in degrees.

Coded locations can also be gathered into **coded location bins** at
a coarser resolution, in order to partition a large dataset into more
manageable chunks:

![A chart showing an example of coded location binning over a New Zealand
grid at 0.1 and 0.5 degrees](../../images/location_binning.png)
"""


import decimal
import warnings
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from nzshm_common.constants import DEFAULT_RESOLUTION
from nzshm_common.location.types import LatLon


@dataclass(init=False, unsafe_hash=True)
class CodedLocation:
    """A location resolved to the nearest point on a grid with given resolution (degrees).

    Refer to [https://stackoverflow.com/a/28750072](https://stackoverflow.com/a/28750072)
    for the techniques used here to calculate decimal places.
    """

    lat: float = field(hash=True)
    lon: float = field(hash=True)
    resolution: float = field(hash=True)

    def __init__(self, lat: float, lon: float, resolution: float) -> None:
        """
        Create a CodedLocation instance.

        Arguments:
            lat: latitude
            lon: longitude
            resolution: the resolution used to resolve the location
        """
        assert 0 < resolution < 180, "Resolution must be between 0 and 180 degrees."

        self.grid_res = decimal.Decimal(str(resolution).rstrip("0"))
        self.display_places = max(abs(self.grid_res.as_tuple().exponent), 1)  # type: ignore

        div_res = 1 / float(self.grid_res)
        places = abs(decimal.Decimal(div_res).as_tuple().exponent)  # type: ignore

        self.lon = round(lon * div_res, places) / div_res
        self.lat = round(lat * div_res, places) / div_res
        self.resolution = resolution

        self._code = f"{self.lat:.{self.display_places}f}~{self.lon:.{self.display_places}f}"

    def __lt__(self, other: "CodedLocation"):
        """
        Less-than comparator to enable sorting for CodedLocations.

        For arithmetic comparisons we expect:

        * north comes before south
        * when on the same latitude, west comes before east

        Note:
            Coded locations at different resolutions are not considered
            equal. Use `.as_tuple` to check for numerical equivalency.

        Examples:
            >>> origin = CodedLocation(-11.11, 111.11, resolution=0.001)
            >>> origin == CodedLocation(-11.11, 111.11, resolution=0.001)  # Equal
            True
            >>> origin == CodedLocation(-11.11, 111.11, resolution=0.01)  # Wrong resolution
            False

            >>> origin < CodedLocation(-11.12, 111.11, resolution=0.001)  # South of origin
            False
            >>> origin < CodedLocation(-11.10, 111.11, resolution=0.001)  # North of origin
            True
            >>> origin < CodedLocation(-11.11, 111.10, resolution=0.001)  # West of origin
            False
            >>> origin < CodedLocation(-11.11, 111.12, resolution=0.001)  # East of origin
            True

        """
        lat_delta = self.lat - other.lat
        lon_delta = self.lon - other.lon
        if lat_delta > 0:
            return True
        elif lat_delta < 0:
            return False
        elif lon_delta < 0:
            return True
        else:
            return False

    @property
    def as_tuple(self) -> LatLon:
        """
        Convert coded location value to a `LatLon(latitude, longitude)` named tuple.

        Example:
            ```py
            >>> from nzshm_common import location
            >>> location.get_locations(["CHC"])[0]
            CodedLocation(lat=-43.53, lon=172.63, resolution=0.001)
            >>> latitude, longitude = location.get_locations(["CHC"])[0].as_tuple
            >>> latitude
            -43.53
            ```
        """
        return LatLon(self.lat, self.lon)

    @property
    def code(self) -> str:
        """
        The string code for the location expressed as "latitude~longitude".

        String codes are padded to a number of decimal places appropriate
        to the resolution property of the coded location.

        Examples:
            >>> from nzshm_common import location
            >>> location.get_locations(["CHC"])[0].code
            '-43.530~172.630'
            >>> location.get_locations(["CHC"], resolution=0.1)[0].code
            '-43.5~172.6'

        """
        return self._code

    @classmethod
    def from_tuple(cls, location: LatLon, resolution: float = DEFAULT_RESOLUTION) -> "CodedLocation":
        """
        Create a `CodedLocation` from a tuple.

        Parameters:
            location: a structure containing a latitude and longitude, in that order
            resolution: coordinate resolution in degrees

        Examples:
            Convert a single location:
            >>> from nzshm_common import CodedLocation
            >>> CodedLocation.from_tuple((-36.87, 174.77))
            CodedLocation(lat=-36.87, lon=174.77, resolution=0.001)

            >>> from nzshm_common import LatLon
            >>> CodedLocation.from_tuple(LatLon(-36.87, 174.77))
            CodedLocation(lat=-36.87, lon=174.77, resolution=0.001)

            Convert a list of locations:
            >>> location_list = [(-36.111, 174.111), (-36.222, 174.222)]
            >>> list(map(CodedLocation.from_tuple, location_list))
            [
                CodedLocation(lat=-36.111, lon=174.111, resolution=0.001),
                CodedLocation(lat=-36.222, lon=174.222, resolution=0.001)
            ]

            Convert a list of locations with a custom resolution:
            >>> from functools import partial
            >>> lo_res = partial(CodedLocation.from_tuple, resolution=0.1)
            >>> list(map(lo_res, location_list))
            [
                CodedLocation(lat=-36.1, lon=174.1, resolution=0.1),
                CodedLocation(lat=-36.2, lon=174.2, resolution=0.1)
            ]
        """
        return CodedLocation(lat=location[0], lon=location[1], resolution=resolution)

    def resample(self, resolution: float) -> "CodedLocation":
        """
        Create a resampled CodedLocation with a finer resolution.

        This operation will not recover precision lost in previous downsampling.

        Examples:
            ```py
            >>> loc
            CodedLocation(lat=-36.1, lon=174.1, resolution=0.1)
            >>> loc.resample(0.01)
            CodedLocation(lat=-36.1, lon=174.1, resolution=0.01)
            >>> loc.resample(0.01).code
            '-36.10~174.10'
            ```
        """
        return self.downsample(resolution)

    def downsample(self, resolution: float) -> "CodedLocation":
        """
        Create a downsampled CodedLocation with a coarser resolution.

        Examples:
            ```py
            >>> from nzshm_common import location
            >>> loc_akl = location.get_locations(["AKL"])[0]
            >>> loc_akl
            CodedLocation(lat=-36.87, lon=174.77, resolution=0.001)
            >>> loc_akl.downsample(0.1)
            CodedLocation(lat=-36.9, lon=174.8, resolution=0.1)
            >>> loc_akl.downsample(0.5)
            CodedLocation(lat=-37.0, lon=175.0, resolution=0.5)
            ```
        """
        return CodedLocation(lat=self.lat, lon=self.lon, resolution=resolution)


class CodedLocationBin:
    """
    A collection of CodedLocation values, gathered into bins at a coarser resolution.
    """

    reference_point: CodedLocation
    locations: List[CodedLocation]
    bin_resolution: float

    def __init__(
        self, reference_point: CodedLocation, bin_resolution: float, locations: Optional[Iterable[CodedLocation]] = None
    ):
        """
        Create a CodedLocationBin instance.

        Arguments:
            reference_point: the downsampled coordinate use as a reference point for the collection.
            bin_resolution: the coarser-level resolution of the bin
            locations: a collection of CodedLocation values
        """
        self.reference_point = reference_point
        self.bin_resolution = bin_resolution

        if locations is not None:
            self.locations: List[CodedLocation] = list(locations)
        else:
            self.locations: List[CodedLocation] = list()

    def __iter__(self):
        """Iterate over the location collection."""
        return self.locations.__iter__()

    def __len__(self):
        """Count the number of locations in the bin."""
        return len(self.locations)

    def __repr__(self):
        plural = "" if len(self.locations) == 1 else "s"
        count = len(self)
        code = self.reference_point.code
        resolution = self.bin_resolution

        return f"CodedLocationBin({count} location{plural} near {code} below resolution {resolution})"

    @property
    def code(self) -> str:
        """
        The string code for the reference point, expressed as "latitude~longitude".

        This can be used as a unique key when accessing a dictionary of bins.

        Example:
            ```py
            >>> my_bin = list(bins.values())[0]
            >>> my_bin
            CodedLocationBin(12 locations near -37.0~175.0 below resolution 0.5)
            >>> bins[my_bin.code] == my_bin
            True
            ```
        """
        return self.reference_point.code


def bin_locations(
    locations: Iterable[CodedLocation], at_resolution: float, sort_bins: bool = True
) -> OrderedDict[str, CodedLocationBin]:
    """
    Collect CodedLocations into a dictionary of bins at a coarser resolution.

    Bin selection is based on the `CodedLocation.downsample` method, reducing
    coordinate precision.

    Arguments:
        locations: a collection of CodedLocations at a finer resolution
        at_resolution: the resolution used when creating CodedLocationBins
        sort_bins: whether to sort the bins and their members

    Returns:
        an ordered dictionary of bins, keyed on the `CodedLocationBin.reference_point.code` property.

    Examples:
        >>> from nzshm_common import grids, location
        >>> grid_locs = grids.get_location_grid('NZ_0_1_NB_1_1', resolution=0.1)
        >>> grid_bins = bin_locations(grid_locs, at_resolution=0.5)
        >>> for location_bin in grid_bins:
        ...     for coded_loc in bin:
        ...         # Do a thing

        To preserve location ordering:
        >>> grid_bins = bin_locations(grid_locs, 0.25, sort_bins=False)
    """
    bin_dict: OrderedDict[str, CodedLocationBin] = OrderedDict()

    max_resolution = 0.0

    for location in locations:
        max_resolution = max(max_resolution, location.resolution)
        coded_loc = location.downsample(at_resolution)
        bin_code = coded_loc._code
        if bin_code not in bin_dict:
            bin_dict[bin_code] = CodedLocationBin(coded_loc, at_resolution, [])

        bin_dict[bin_code].locations.append(location)

    if max_resolution > at_resolution:
        warn_msg = (
            f"Found locations up to {max_resolution} degree resolution. "
            + f"Binning expected to downsample to a {at_resolution} degree resolution."
        )
        warnings.warn(warn_msg, stacklevel=2)

    if sort_bins:
        # Sort the bins themselves.
        bin_dict = OrderedDict(sorted(bin_dict.items(), key=lambda item: item[1].reference_point))

        # Sort CodedLocations within each bin.
        for location_bin in bin_dict.values():
            location_bin.locations.sort()

    return bin_dict



import pytest

from nzshm_common.location.location import (
    DEFAULT_RESOLUTION,
    LOCATION_LISTS,
    get_location_list,
    get_location_list_names,
    get_locations,
)


def test_single_source():
    location_list = get_location_list(["NZ"])
    assert len(location_list) == len(LOCATION_LISTS["NZ"]["locations"]), "Should match count for NZ locations"

    location_list_codes = [loc.code for loc in location_list]
    assert get_locations(["AKL"])[0].code in location_list_codes, "Should find Auckland"
    assert get_locations(["DUD"])[0].code in location_list_codes, "Should find Dunedin"


def test_multiple_sources():
    location_list = get_location_list(["NZ", "SRWG214"])
    merged_locations = set(LOCATION_LISTS["NZ"]["locations"] + LOCATION_LISTS["SRWG214"]["locations"])
    assert len(location_list) == len(merged_locations), "List length equalling merge of two lists"


def test_source_overlaps():
    """
    Ensure that locations that appear in multiple lists are not duplicated.
    """
    nznz_list = get_location_list(["NZ", "NZ"])
    assert len(nznz_list) == len(LOCATION_LISTS["NZ"]["locations"]), "Should match count for single NZ location list"

    masterton = get_locations(["MRO"])[0]
    # Turning off sort_locations to make the test marginally faster for larger collections.
    nz_list = get_location_list(["NZ"], sort_locations=False)
    all_list = get_location_list(["ALL"], sort_locations=False)
    nz_all_list = get_location_list(["NZ", "ALL"], sort_locations=False)

    assert nz_list.count(masterton) == 1, "Should find Masterton in New Zealand once"
    assert all_list.count(masterton) == 1, "Should find Masterton in All once"
    assert nz_all_list.count(masterton) == 1, "Should find Masterton in combined list once"


def test_missing_source():
    with pytest.raises(KeyError):
        get_location_list(["unknown"])

    with pytest.raises(KeyError):
        get_location_list(["NZ", "unknown"])

    location_list = get_location_list([])
    assert len(location_list) == 0, "Should be an empty list"


def test_resolution_override():
    custom_resolution = 0.1
    assert DEFAULT_RESOLUTION != custom_resolution, "Tested resolutions should be different"

    location_list = get_location_list(["NZ"], resolution=custom_resolution)
    assert len(location_list) == len(LOCATION_LISTS["NZ"]["locations"]), "Should match count for NZ locations"

    location_list_codes = [loc.code for loc in location_list]
    assert (
        get_locations(["AKL"])[0].code not in location_list_codes
    ), "Should not find Auckland code at default resolution"
    assert (
        get_locations(["AKL"], resolution=custom_resolution)[0].code in location_list_codes
    ), "Should find Auckland code at custom resolution"


def test_names():
    name_list = get_location_list_names()
    assert "NZ" in name_list, "Should contain NZ location list"
    assert "SRWG214" in name_list, "Should contain SRWG214 location list"
    assert "NZ_0_1_NB_1_0" not in name_list, "Should not include grid names"



import base64
import csv
import io
import zipfile
from typing import List

from nzshm_common.location.types import LatLon


def latlon_from_base64_zip(b64_data: str) -> List[LatLon]:
    result = []
    byte_data = base64.b64decode(b64_data)
    data = io.BytesIO(byte_data)
    with zipfile.ZipFile(data) as fz:
        plain_bytes = io.BytesIO(fz.read("data"))
        reader = csv.reader(io.StringIO(plain_bytes.getvalue().decode("utf-8", "ignore")))
        for row in reader:
            result.append(LatLon(float(row[1]), float(row[0])))
    return result



import warnings
from collections import namedtuple
from enum import Enum
from functools import partial
from typing import Iterable, List, Optional

from nzshm_common.grids.nz_0_1_nb_1_v0 import NZ01nb1v0
from nzshm_common.grids.nz_0_1_nb_1_v1 import NZ01nb1v1
from nzshm_common.grids.nz_0_2_nb_1_1 import NZ_0_2_nb_1_1
from nzshm_common.grids.wlg_0_01_nb_1_1 import WLG_0_01_nb_1_1
from nzshm_common.grids.wlg_0_05_nb_1_1 import WLG_0_05_nb_1_1
from nzshm_common.location.coded_location import CodedLocation
from nzshm_common.location.types import LatLon

RegionGridEntry = namedtuple("RegionGridEntry", "region_name resolution neighbours grid version")


class RegionGrid(Enum):
    """
    An enumerated collection of region grids defined in this module.
    """

    NZ_0_1_NB_1_0 = RegionGridEntry(region_name="NZ", resolution=0.1, neighbours=1, grid=NZ01nb1v0(), version=0)
    NZ_0_1_NB_1_1 = RegionGridEntry(region_name="NZ", resolution=0.1, neighbours=1, grid=NZ01nb1v1(), version=1)
    NZ_0_2_NB_1_1 = RegionGridEntry(region_name="NZ", resolution=0.2, neighbours=1, grid=NZ_0_2_nb_1_1(), version=1)

    WLG_0_01_nb_1_1 = RegionGridEntry(
        region_name="WLG",
        resolution=0.01,
        neighbours=1,
        grid=WLG_0_01_nb_1_1(),
        version=1,
    )
    WLG_0_05_nb_1_1 = RegionGridEntry(
        region_name="WLG",
        resolution=0.05,
        neighbours=1,
        grid=WLG_0_05_nb_1_1(),
        version=1,
    )

    def __init__(self, region_name, resolution, neighbours, grid, version):
        self.region_name = region_name
        self.resolution = resolution
        self.neighbours = neighbours
        self.grid = grid

    def load(self):
        return self.grid.load()


def load_grid(grid_name: str) -> List[LatLon]:
    """
    Load values from a region grid as `LatLon` pairs.

    Examples:
        >>> from nzshm_common import grids
        >>> grids.load_grid("NZ_0_1_NB_1_0")
        [
            LatLon(latitude=-46.1, longitude=166.4),
            LatLon(latitude=-46.0, longitude=166.4),
            LatLon(latitude=-45.9, longitude=166.4),
            ...
        ]
    """
    return RegionGrid[grid_name].load()


def get_location_grid(location_grid_name: str, resolution: Optional[float] = None) -> Iterable[CodedLocation]:
    """
    Get all coded locations within a grid.

    Note:
        When downsampling to a lower resolution, duplicate location values will be removed.

    Parameters:
        location_grid_name: A valid member key from RegionGrid (e.g. "NZ_0_1_NB_1_1")
        resolution: The resolution of the CodedLocation values generated.
            Defaults to the native RegionGrid resolution value.

    Returns:
        A list of coded locations with the specified resolution.

    Examples:
        >>> from nzshm_common import grids
        >>> grids.get_location_grid("NZ_0_1_NB_1_0")
        [
            CodedLocation(lat=-46.1, lon=166.4, resolution=0.1),
            CodedLocation(lat=-46.0, lon=166.4, resolution=0.1)
            ...
        ]
    """
    if not resolution:
        resolution = RegionGrid[location_grid_name].resolution
    elif resolution > RegionGrid[location_grid_name].resolution:
        warn_msg = "The requested resolution is lower than the grid resolution and will result in fewer points."
        warnings.warn(warn_msg, stacklevel=2)

    grid_values = load_grid(location_grid_name)
    coded_at_resolution = partial(CodedLocation.from_tuple, resolution=resolution)
    # Remove duplicate coordinates from collection, preserving order.
    location_list = []
    for loc in map(coded_at_resolution, grid_values):
        if loc not in location_list:
            location_list.append(loc)

    return location_list


def get_location_grid_names() -> Iterable[str]:
    """
    Return a collection of of valid region grids.

    Returns:
        member names from the RegionGrid Enum class.

    Examples:
        >>> from nzshm_common import grids
        >>> grids.get_location_grid_names()
        dict_keys([
            'NZ_0_1_NB_1_0',
            'NZ_0_1_NB_1_1',
            'NZ_0_2_NB_1_1',
            'WLG_0_01_nb_1_1',
            'WLG_0_05_nb_1_1'
        ])
    """

    return RegionGrid.__members__.keys()



from collections import OrderedDict
from pathlib import Path

import pytest

from nzshm_common.location.coded_location import CodedLocationBin, bin_locations
from nzshm_common.location.location import get_locations

# Fixture is a subset of the NZ_0_1_NB_1_0 dataset to improve execution time.
# This set includes the first and last bin at 0.5 degrees, and another 18
# randomly sampled bins in between.
GRID_LOCS_FILEPATH = str(Path(__file__).parent / 'fixtures' / 'grid_loc_subset.csv')

GRID_LOCS = get_locations([GRID_LOCS_FILEPATH])
GRID_LOCS_EXPECTED_BIN_COUNT_05 = 20


def test_coded_location_bin_empty():
    single_bin = CodedLocationBin(GRID_LOCS, bin_resolution=1)
    assert len(single_bin.locations) == 0, "A bin can be instantiated with no locations in it"


def test_coded_location_binning_nosort():
    """
    A collection of bins can be left unsorted.
    """
    grid_bins = bin_locations(GRID_LOCS, at_resolution=0.5, sort_bins=False)
    assert isinstance(grid_bins, OrderedDict), "Bin collection should be ordered dictionary"
    assert len(grid_bins) == GRID_LOCS_EXPECTED_BIN_COUNT_05, "Bin count should match expectation"

    first_loc = GRID_LOCS[0]
    expected_first_bin_key = first_loc.downsample(0.5).code
    first_bin_key = list(grid_bins.keys())[0]
    assert first_bin_key == expected_first_bin_key, "First bin should match the downsampled code of the first location"

    first_bin = grid_bins[first_bin_key]
    assert isinstance(first_bin, CodedLocationBin), "Should be CodedLocationBin"
    assert len(first_bin) == len(first_bin.locations), "__len__ should reflect locations list length"
    assert first_bin.locations[0] == first_loc, "Unsorted, first location should be first item in the first bin"

    all_bin_counts = [len(grid_bin) for grid_bin in grid_bins.values()]
    assert sum(all_bin_counts) == len(GRID_LOCS), "Bin counts should sum to total location count"

    assert first_bin.locations == [loc for loc in first_bin], "Bin __iter__ should match its locations list"


def test_coded_location_binning_sort():
    """
    A collection of bins should be sorted geographically by default.
    """
    grid_bins = bin_locations(GRID_LOCS, at_resolution=0.5)

    assert isinstance(grid_bins, OrderedDict), "Bin collection should be ordered dictionary"
    assert len(grid_bins) == GRID_LOCS_EXPECTED_BIN_COUNT_05, "Bin count should match expectation"

    bin_list = list(grid_bins.values())
    min_bin = min(bin_list, key=lambda b: b.reference_point)
    max_bin = max(bin_list, key=lambda b: b.reference_point)

    assert min_bin == bin_list[0], "'Smallest' bin should be first"
    assert grid_bins[min_bin.code] == min_bin
    assert max_bin == bin_list[-1], "'Largest' bin should be last"
    assert grid_bins[max_bin.code] == max_bin

    assert min_bin.locations == sorted(min_bin.locations), "Bin locations should be sorted"
    assert min_bin.locations == sorted([loc for loc in min_bin])

    print()
    print("Northmost bin:", min_bin)
    print("Southmost bin:", max_bin)


def test_coded_location_binning_wrong_resolution():
    """
    If binning encounters coded locations at a coarser resolution than the
    bins, warn the user.
    """
    locs = ["-34.5~172.5", "-47.5~167.5"]
    coded_locs = get_locations(locs, resolution=0.5)

    expected_message = (
        "Found locations up to 0.5 degree resolution. Binning expected to downsample to a 0.01 degree resolution."
    )

    with pytest.warns(UserWarning) as warnings:
        bins = bin_locations(coded_locs, at_resolution=0.01)

    assert len(bins) == len(locs), "Should be one bin per location"
    assert "-34.50~172.50" in bins, "Should find first location resampled"
    assert "-47.50~167.50" in bins, "Should find first location resampled"

    assert len(warnings) == 1, "Should only be warned once"
    assert str(warnings[0].message) == expected_message, "Should have matched expected message"
    assert warnings[0].filename == __file__, "Should have been attributed to this caller"

    print()
    print(warnings[0].message)



from nzshm_common.grids.io import latlon_from_base64_zip


class NZ01nb1v1:
    nz = None

    def load(self):
        if self.nz is None:
            self.nz = latlon_from_base64_zip(data)
        return self.nz


data = (
    "UEsDBD8AAgAOADxX81S60g40OQgAAPm9AAAEAAAAZGF0YQkEBQBdAACAAAAY"
    "jaxcNpfygoevll45+3II7cxm7iUwhrlQ97Th69ymR0q/sK97FS9DzGoX6dMx"
    "F3QtTVZkGQ4y2pyR5wPopKJDZGCFRWosJskYY90YAjRx6UCc6DnfRtzdgpFN"
    "Md9WfbMNi0hAPGKv6ZAuYtqStgsEL8OjuLqMvUplWbx01KOTEc++On40nzQ8"
    "hsg5oZcaDQNyzeWNLIrTULom7u743FB5hjgk7Bd/GN6fcINDdbQvopAO/hC9"
    "u1YiEl/KwwAtt8nXYGnhaej6Qdv31fBEQK9X+MZA51xNvovdXCgmqZ+0fo0z"
    "lHrWw01JS+KQJxNm5hyTVHguYTyF9g2866ZTS/z5x2wYHuUOMO4u11Ba6CdV"
    "2n/NTn+ekwGqrXujDWdjDjhx2mZSlMN/pvHqFAazB2JvRBGXDCUbb3mlE61N"
    "3Nf44MB1LA/5GTJf/GHpmQrDl5c7ibVWkDukRASl6jRLfEYbV5lj2Qfr4Zs9"
    "UXC/Hb8xK0MEUkHfixlAsmA2hZwppNXlaTV60ykLkiSrNcJh67ooLgfqPTll"
    "U3egTG124E+FPF1SLs3kR7JO4YNSAzbk951XzFok9ulvILmPPGRD64Q7hTQU"
    "SkdE77OW88R+3Y8tGhmcKXfKM65k6wxcXLntRCVVgpF1eu8amNaXlIXPHS3y"
    "3GoYNjWii1kqM4ohdM+AwUOTOHwxn1tNVBIPhlCfbZ+W7lJ/O5Jq7uZ6aJHh"
    "TTIrO7cKwR29lc6CppA5n5ubQdiMU1O5r9diqlAT6ANgX7tusX5jV0dmZFBh"
    "G+Zu9loM6R8KXu9w4qJ+z8WmzYL18m2qWzPBeDbexYmtOtL0jK88eYv8BY22"
    "0A5uzld9PhV2WFtFQZqHePcNr0dWtNE/thzYsPj+QbPIywXeg0ihkhch7rnp"
    "02RFHx558ex2k0u8O5Hl8mccI3gsranGEPuj6NPf153FdYYyZp/1feGLJ0Bk"
    "WUGeEt0BtSTX8CwW3q0geupvbP7al/tkcGgMLi5+sZlAEVNu5MKtbKwIGFHG"
    "Z5Z60w2qeDpp23hqwRIBxRKKNA5opBkAAjHKI6LDrGVjtmY+jbXT1BYGRSut"
    "mUDInN3t9+e7r47gQ/re2pEbrl65JFZJHG0580cnqxus+nkxd7rAedcjmu3q"
    "fPSwZMNIlzfra1IbAlzm9XyGwB+izEN2E2QJ1gTzwVvBxzWMjVjYUKB460vW"
    "4KY9z7b6BX+H2qneP8iYUWNM5pfLlEtsB76Qj2S8udgGPmAguRS3x4dv7oMI"
    "8/mH/6KHVydaeUF8oYNnwnCsygjJcsdZpvsMMEGgtkiTAJjLpvL4ZBra7A3g"
    "v6aGLkg2uPVkXK7PDET+Vl5/CW04BtwqtosToqMUNNYPJMaALWuKqLGZCMVw"
    "Hy6hRooiRiGZUd/uUbxnIrlXPY0NLY20sw+MVAr39cPU40qMNQh21Ge2du9I"
    "1zU7wTIP2f3cWbNbnyXH+3R+VouMV9OOiaq6ci8zx34Zzufh8WHU5qKzDU6t"
    "yAmmyKw0Hu+jOqOwDYaS+rWqPoPpg7hBXZTsEJ6z/yAFrN2NlymXQ0ko7vIB"
    "QPcxyLJdi+qLybg+JoTCWJICXvVyLYFDYLsTuHVsax4WmI2XtIHNMx62ViwW"
    "eO5Izc9eafxiIdqpsU/HBUSmlSfKOpFHd88B2IXORZ6FnWMURvxQRDvYpxPd"
    "AlwDCSaE4+SDdYFkVkEI8nSUwARhw6QnHiD30dUOap+5EwptZ9VB3MXaA5j8"
    "a5FCCdFMK02Y5JtMa2LQqVBcykBcr0XFzPcXElumKRD8LDDITPBnP3vYt4Y8"
    "AsJBStzQGxk+WJi3VvO/r3EocN6XYcpRJYJ+2oZRJk/HXXEmzKYMzMKLjaK8"
    "mR/EJJTGCFEzabgTBTJANWs9kpK1CF6Yqnhx/dvyYAkuTJfEnfFt2Y0eseHm"
    "e12Dk9t5tBvi68XWkWkD+5jA7C3GINmVvrV4aPiZoBOwjxNcomd32hoc8WdL"
    "AGNWvfyOKiA/oQywA0H1f8km+Vgiqq+GYatXkI67ReL4rrCKmQd8HwtW4O81"
    "ygAAQsonaTSHo6PW4P5WJOBDrIytHDXw7MZmMKpa17tHdseH73V17o7CwowO"
    "Ihg1hEnczMRouwlMaA7dbWxm/VnPT95kXxgDU+L3FbET0F6GMMuLpD+s82LQ"
    "wbK3ylDXv3S71lzsNCAja0xMWGEpdKh1b+l2CoQbALRkCZxgR7t0ms+h/J9w"
    "s3tYRhrgYZNa7NznAv6Ok6DmaRepZau0r7WvWVl5obBZ3NFyAmdmsr5lITlI"
    "qjX+QgbpVvrMs4VcnWR7IYYHjcE8lL9gK3jGG1yF80++HGRywLihfaZySpoM"
    "kugLkcRmrEt8lGoYXONHPbvzJe2RaFtnuuTICM6Uc7SdZNNrgP7H9DKhTBcE"
    "hkIKT5l/FgLhuqvgSTUezf9GHfu+oSKnBZTxO4+qYv8GQ3qT/KUB/VMZ8s48"
    "hTMq7sCEPM4LWCpwx6V/FBFG1lF+/jm9224+RxOksdAFqRnRFq3YeTgXqc+m"
    "x/agLfsopBtHYrZNgK7yc0jJuk99touaXzly21H1hcHujpnrh21b7XBr5iex"
    "E7h7dGh/nm0r/bU9a5x08w8r98PLTeL5yMtwzEUY9NklzpzAG2p84M2hjhMz"
    "YhwbdIgn+lCclTyxKuTgtrn03pe3NZqZdNcrvo6M1bRRmqN1XlDacFzTM96W"
    "qaPaN/WIKM1x+TypmhIK2Td2LBBhZ7D9Pl0kxrn1G7ctmNTsQwbl4uC/tu7N"
    "flRQdE4Ii44DI/NEbETngmjBv/9xWDEAUEsBAj8DPwACAA4APFfzVLrSDjQ5"
    "CAAA+b0AAAQAAAAAAAAAAAAAAIABAAAAAGRhdGFQSwUGAAAAAAEAAQAyAAAA"
    "WwgAAAAA"
)



import base64
import io
import zipfile
from pathlib import Path, PurePath
from typing import Union


def compress_path(source_path: str, output_zip: Union[str, PurePath], compression=zipfile.ZIP_DEFLATED) -> str:
    """Zip contents of source path and return the archive path.

    Handles both single file and a folder.
    """
    with zipfile.ZipFile(output_zip, "w", compression=compression) as zf:
        if Path(source_path).is_file():
            zf.write(source_path, PurePath(source_path).name)
        else:
            for filename in Path(source_path).rglob("*"):
                zf.write(filename, arcname=str(Path(filename).relative_to(source_path)))
    return str(output_zip)


def compress_string(data: str, compression=zipfile.ZIP_LZMA) -> str:
    """Use compression to pack this string into a much smaller string."""
    compressed = io.BytesIO()
    with zipfile.ZipFile(compressed, "w", compression=compression) as zf:
        zf.writestr("0", data)
    compressed.seek(0)
    b64 = base64.b64encode(compressed.read())
    return b64.decode("ascii")


def decompress_string(compressed: str) -> str:
    """Decompress a ZIP compressed string to the original string."""
    base64_bytes = compressed.encode("ascii")
    message_bytes = base64.b64decode(base64_bytes)
    with zipfile.ZipFile(io.BytesIO(message_bytes)) as zf:
        b64 = io.BytesIO(zf.read("0"))
    b64.seek(0)
    return b64.read().decode("utf-8")



import random
import unittest

import pytest

from nzshm_common import CodedLocation, LatLon
from nzshm_common.constants import DEFAULT_RESOLUTION
from nzshm_common.grids.region_grid import load_grid
from tests.helpers import does_not_raise

GRID_02 = load_grid('NZ_0_2_NB_1_1')
LOCS = [CodedLocation(loc[0], loc[1], 0.001) for loc in GRID_02[20:50]]  # type: ignore


def test_coded_location_is_hashable():
    c = CodedLocation(-45.2, 175.2, 0.1)
    s = set()
    s.add(c)
    assert c in s


class CodedLocationResampling(unittest.TestCase):
    def test_get_nearest_hazard_for_an_arbitrary_location(self):
        gridloc = random.choice(LOCS)
        print(f'gridloc {gridloc}')

        off_lat = round(gridloc.lat + random.randint(0, 9) * 0.01, 3)
        off_lon = round(gridloc.lon + random.randint(0, 9) * 0.01, 3)
        somewhere_off_grid = CodedLocation(off_lat, off_lon, 0.001)

        nearest_grid = somewhere_off_grid.downsample(0.2)

        print(f'somewhere_off_grid {somewhere_off_grid}')
        print(f'nearest_grid {nearest_grid}')

        self.assertEqual(gridloc, nearest_grid.resample(0.001))
        self.assertEqual(gridloc, nearest_grid.downsample(0.001))
        self.assertEqual(gridloc.code, nearest_grid.downsample(0.001).code)

        self.assertEqual(gridloc, CodedLocation(nearest_grid.lat, nearest_grid.lon, 0.001))
        self.assertTrue(CodedLocation(nearest_grid.lat, nearest_grid.lon, 0.001) in LOCS)


oh_point_five_expected = [
    (-45.27, 171.1, '-45.5~171.0'),
    (-45.23, 171.1, '-45.0~171.0'),
    (-45.27, 171.4, '-45.5~171.5'),
    (-45.27, 171.8, '-45.5~172.0'),
    (-41.3, 174.783, '-41.5~175.0'),  # WLG
]


def test_as_tuple():
    c = CodedLocation(-45.2, 175.2, 0.1)

    assert isinstance(c.as_tuple, LatLon)
    assert c.as_tuple.latitude == -45.2
    assert c.as_tuple.longitude == 175.2
    assert c.as_tuple[0] == -45.2
    assert c.as_tuple[1] == 175.2


@pytest.mark.parametrize("lat,lon,expected", oh_point_five_expected)
def test_coded_location_equality(lat, lon, expected):
    c0 = CodedLocation(lat, lon, 0.5)
    c1 = CodedLocation(lat, lon, 0.5)
    assert c0 == c1


def test_coded_location_from_tuple():
    coded_loc = CodedLocation.from_tuple(LatLon(latitude=-45.27, longitude=171.14))
    assert isinstance(coded_loc, CodedLocation), "Return type should be CodedLocation"
    assert coded_loc.resolution == DEFAULT_RESOLUTION, "Should have default resolution"
    assert coded_loc.lat == -45.27, "Latitude should match"
    assert coded_loc.lon == 171.14, "Longitude should match"

    # A naked (latitude, longitude) tuple with the same values should work also.
    coded_loc_lores = CodedLocation.from_tuple((-45.27, 171.14), resolution=0.1)
    assert coded_loc_lores.resolution == 0.1, "Should have lowered resolution"
    assert coded_loc_lores.lat == -45.3, "Should have rounded latitude"
    assert coded_loc_lores.lon == 171.1, "Should have rounded longitude"


@pytest.mark.parametrize(
    "lat,lon,is_before,is_code_before,description",
    [
        (-45.1, +171.3, 1, 1, "0.1 North 0.1 West"),
        (-45.1, +171.4, 1, 1, "0.1 North"),
        (-45.1, +171.5, 1, 1, "0.1 North 0.1 East"),
        (-45.2, +171.3, 1, 1, "0.1 West"),
        (-45.2, +171.4, 0, 0, "Reference (in Southern/Eastern Hemispheres)"),
        (-45.2, +171.5, 0, 0, "0.1 East"),
        (-45.3, +171.3, 0, 0, "0.1 South 0.1 West"),
        (-45.3, +171.4, 0, 0, "0.1 South"),
        (-45.3, +171.5, 0, 0, "0.1 South 0.1 East"),
        (+45.2, -171.4, 1, 0, "Northern / Western Hemispheres (code sort inverts lat)"),
        (+45.2, +171.4, 1, 0, "Northern / Eastern Hemispheres (code sort inverts lat)"),
        (-45.2, -171.4, 1, 1, "Southern / Western Hemispheres"),
    ],
)
def test_coded_location_ordering(lat, lon, is_before, is_code_before, description):
    """
    Characterising differences in sorting arithmetically by latitude then
    longitude, versus alphanumeric .code sorting.

    Alpha sorting works for New Zealand because we have a negative latitude,
    but would sort in the opposite direction for the northern hemisphere.

    For arithmetic comparisons we expect:

    - North before Sorth, or
    - West before East when on the same latitude
    """
    reference_point = CodedLocation(-45.24, 171.4, 0.1)
    is_before = bool(is_before)
    is_code_before = bool(is_code_before)

    target = CodedLocation(lat, lon, 0.1)
    assert (target < reference_point) == is_before, f"Comparison expected: {is_before}"
    assert (target.code < reference_point.code) == is_code_before, f"Code comparison expected: {is_code_before}"


@pytest.mark.parametrize("lat,lon,expected", oh_point_five_expected)
def test_downsample_default_oh_point_five_no_downsampling_required(lat, lon, expected):
    print(f"lat {lat} lon {lon} -> {expected}")
    assert CodedLocation(lat, lon, 0.5).code == expected


@pytest.mark.parametrize("lat,lon,expected", oh_point_five_expected)
def test_downsample_default_oh_point_five(lat, lon, expected):
    print(f"lat {lat} lon {lon} -> {expected}")
    c = CodedLocation(lat, lon, 0.5)
    assert c.downsample(0.5).code == expected


@pytest.mark.parametrize(
    "lat,lon,expected",
    [
        (-45.27, 171.1, '-45.0~171.0'),
        (-45.23, 171.1, '-45.0~171.0'),
        (-45.77, 171.4, '-46.0~171.0'),
        (-45.27, 171.8, '-45.0~172.0'),
        (-41.3, 174.78, '-41.0~175.0'),  # WLG
    ],
)
def test_downsample_one_point_oh(lat, lon, expected):
    c = CodedLocation(lat, lon, 1.0)
    assert c.downsample(1.0).code == expected


@pytest.mark.parametrize(
    "lat,lon,expected",
    [
        (-45.27, 171.1, '-45.3~171.1'),
        (-45.239, 171.13, '-45.2~171.1'),
        (-45.27, 171.4, '-45.3~171.4'),
        (-45.27, 171.8, '-45.3~171.8'),
        (-41.333, 174.78, '-41.3~174.8'),  # WLG
    ],
)
def test_downsample_oh_point_one(lat, lon, expected):
    c = CodedLocation(lat, lon, 0.1)
    assert c.downsample(0.1).code == expected


@pytest.mark.parametrize(
    "lat,lon,expected",
    [
        (-45.27, 171.111, '-45.25~171.10'),
        (-45.239, 171.73, '-45.25~171.75'),
        (45.126, 171.4, '45.15~171.40'),
        (-45.27, 171.03, '-45.25~171.05'),
        (-41.333, 174.78, '-41.35~174.80'),  # WLG
    ],
)
def test_downsample_oh_point_oh_five(lat, lon, expected):
    c = CodedLocation(lat, lon, 0.05)
    assert c.downsample(0.05).code == expected


@pytest.mark.parametrize(
    "resolution, expectation",
    [
        (-0.1, pytest.raises(AssertionError)),
        (0.0, pytest.raises(AssertionError)),
        (0.05, does_not_raise()),
        (0.1, does_not_raise()),
        (150, does_not_raise()),
        (180, pytest.raises(AssertionError)),
    ],
)
def test_resolution_bounds(resolution, expectation):
    """Ensure invalid resolutions throw an assertion error before calculating."""
    with expectation:
        CodedLocation(-41.333, 174.78, resolution)



import os
import random
import string
import tempfile
import zipfile
from pathlib import Path

from nzshm_common.util import compress_path, compress_string, decompress_string


def test_string_compression_round_trip():
    random_input = ''.join(random.choices(string.ascii_uppercase + string.digits, k=1028))
    round_tripped = decompress_string(compress_string(random_input))
    assert random_input == round_tripped


def test_string_compression_round_trip_unicode():
    unicode = "āĀēĒīĪōŌūŪ Le cœur déçu mais l'âme plutôt naïve. ‘Single’ and “double” quotes, “curly” apostrophes."
    round_tripped = decompress_string(compress_string(unicode))
    assert unicode == round_tripped


def test_compress_path_round_trip():
    random_content_A = ''.join(random.choices(string.ascii_uppercase + string.digits, k=1024))

    # create a temporary directory using the context manager
    tmpdirname = tempfile.mkdtemp()
    assert Path(tmpdirname).exists()

    fileA = None
    try:
        with tempfile.NamedTemporaryFile(prefix='Random_data_', dir=tmpdirname, mode='w', delete=False) as fileA:
            # ref https://stackoverflow.com/a/49868505
            fileA.write(random_content_A)
            fileA.flush()

            print(fileA.name)
            assert Path(fileA.name).exists()

            archive_file_name = compress_path(fileA.name, Path(tmpdirname, 'test_archive_A.zip'))

            print(f'archive_file_name: {archive_file_name}')
            assert Path(archive_file_name).exists()

            with zipfile.ZipFile(archive_file_name, 'r') as zip_ref:
                for info in zip_ref.infolist():
                    print(info)

                tmp2name = tmpdirname
                with zipfile.ZipFile(archive_file_name, 'r') as zip_ref:
                    # assert fileA.name in zip_ref.namelist()
                    zip_ref.extractall(tmp2name)

                extracted_fileA = Path(tmp2name, fileA.name)

                print(extracted_fileA)
                print(list(Path(tmp2name).glob('*')))
                assert extracted_fileA.exists()

                with open(extracted_fileA, 'r') as newFileA:
                    assert newFileA.read() == random_content_A
    finally:
        if fileA is not None:
            os.remove(fileA.name)



import pytest

import nzshm_common.location.location as location
from nzshm_common.location.coded_location import CodedLocation


def test_location_keys_unique():
    assert len(location.LOCATIONS) == len(set(loc['id'] for loc in location.LOCATIONS))


def test_location_lists():
    assert len(location.LOCATION_LISTS["NZ"]["locations"]) == 36
    assert len(location.LOCATION_LISTS["SRWG214"]["locations"]) == 214
    assert len(location.LOCATION_LISTS["ALL"]["locations"]) == 214 + 36 + 19480
    assert len(location.LOCATION_LISTS["HB"]["locations"]) == 19480


def test_vs30():
    for id in location.LOCATION_LISTS["HB"]["locations"]:
        assert location.LOCATIONS_BY_ID[id].get("vs30")
    assert location.LOCATIONS_BY_ID[f"hb_{2603-2}"]["vs30"] == 150


def test_location_rot():
    rot = location.LOCATIONS_BY_ID['ROT']
    assert rot['name'] == 'Rotorua'


def test_location_pauanui():
    rot = location.LOCATIONS_BY_ID['srg_34']
    assert rot['name'] == 'Pauanui'


def test_hawks_bay():
    hb0 = location.LOCATIONS_BY_ID['hb_0']
    assert hb0['latitude'] == -39.28856686
    assert hb0['longitude'] == 176.1209845
    assert hb0['vs30'] == 1000

    hb19479 = location.LOCATIONS_BY_ID['hb_19479']
    assert hb19479['latitude'] == -38.15959922
    assert hb19479['longitude'] == 178.2035461
    assert hb19479['vs30'] == 1000


def test_rounded_locations():
    def get_lat_lon(id):
        return (location.location_by_id(id)['latitude'], location.location_by_id(id)['longitude'])

    id = 'srg_142'
    assert CodedLocation(*get_lat_lon(id), 0.001).code == "-41.520~173.948"

    id = 'srg_186'
    assert CodedLocation(*get_lat_lon(id), 0.001).code == "-44.379~171.230"


def test_missing_lat_lon_returns_None():
    assert location._lat_lon("missingid") is None, "An unknown ID should return a None"


def test_word_mapping():
    for k, v in location.WORD_MAPPING.items():
        macrons = ['ĀĒĪŌŪāēīōū']
        assert len(k.split()) == 1
        assert len(v.split()) == 1
        assert len(v) == len(k)
        assert all(char not in macrons for char in k)  # the keys should not contain macrons


@pytest.mark.parametrize(
    "name_in,name_out",
    [
        ("Otaki", "Ōtaki"),  # should work with capital letters
        ("Hamama", "Hāmama"),
        ("Kerepēhi", "Kerepēhi"),
        ("Ahititi", "Ahitītī"),
        ("Haukopua Point", "Haukōpua Point"),
        ("Haututerangi", "Hautūterangi"),
        ("Oakura (New Plymouth District)", "Ōakura (New Plymouth District)"),  # handle whole word parts of names
        ("Wellington", "Wellington"),  # not in the name list, should return the input
    ],
)
def test_macron_mapping(name_in, name_out):
    assert location.get_name_with_macrons(name_in) == name_out
